import uuid
import logging
import pymill

from urlparse import urlparse
from django.core.urlresolvers import reverse, resolve
from django.core.exceptions import ImproperlyConfigured

from . import PaymentProcessor

logger = logging.getLogger(__name__)


WEBHOOK_EVENTS = (
    u'chargeback.executed',
    u'refund.created',
    u'refund.succeeded',
    u'refund.failed',
    u'subscription.created',
    u'subscription.updated',
    u'subscription.deleted',
    u'subscription.succeeded',
    u'subscription.failed',
    u'transaction.created',
    u'transaction.succeeded',
    u'transaction.failed',
    u'invoice.available',
    u'payout.transferred',
    u'app.merchant.activated',
    u'app.merchant.deactivated',
    u'app.merchant.rejected',
    u'client.updated',
    u'app.merchant.app.disabled'
)


class Webhook(object):
    def __init__(self, *args, **kwargs):
        self.Pymill = pymill.Pymill
        self.host = PaymentProcessor.get_backend_setting('PAYMILL_WEBHOOK_HOST')
        self.secret = None
        return super(Webhook, self).__init__(*args, **kwargs)

    def set_pymill(self, pymill_class):
        '''IoC for testings'''
        self.Pymill = pymill_class

    def get_webhook(self):
        if self.secret:
            return self.secret

        paymill = self.Pymill(PaymentProcessor.get_backend_setting('PAYMILL_PRIVATE_KEY'))
        webhooks = paymill.get_webhooks()
        host = urlparse(self.host)
        for hook in webhooks:
            url = urlparse(hook.url)
            try:
                match = resolve(url.path)
                if host.scheme == url.scheme and host.netloc == url.netloc and match.url_name == 'abo-paymill-webhook':
                    self.secret = match.kwargs.get('secret', None)
                    return self.secret
            except:
                pass

        return self.secret

    def install_webhook(self):
        paymill = self.Pymill(PaymentProcessor.get_backend_setting('PAYMILL_PRIVATE_KEY'))
        if not self.host:
            raise ImproperlyConfigured('You need to set PAYMILL_WEBHOOK_HOST in your backend settings for paymill')
        secret = uuid.uuid4().hex
        url = '%s%s' % (
            self.host,
            reverse('abo-paymill-webhook', args=[secret, ]))
        logger.info('Registering webhook: %s', url)
        paymill.new_webhook(url, WEBHOOK_EVENTS)
        self.secret = secret

    def init_webhook(self):
        logger.info('Looking for webhook')
        secret = self.get_webhook()
        if not secret:
            logger.warning('Webhook not found, installing')
            self.install_webhook()

    def get_url(self):
        if not self.secret:
            self.get_webhook()
        return self.host + reverse('abo-paymill-webhook', args=[self.secret, ])



from abo.models import BackendSubscription
from .webhooks import WEBHOOK_EVENTS


class EventProcessor():
    """
    Handle the paymill events we received through a webhook

    This will be used by BackendEvent.process()
    """

    def __init__(self, backend_event):
        self.backend_event = backend_event
        self.object = None

    def process(self):
        # TODO also handle errors in message
        event = self.backend_event.message.get('event')

        if not self.backend_event.event_type in WEBHOOK_EVENTS:
            raise NotImplementedError(self.backend_event.event_type)

        event_fn_name = self.backend_event.event_type.replace('.', '_')
        event_fn = getattr(self, "_" + event_fn_name, None)

        if event_fn:
            event_fn(event)

        if self.object:
            self.object.save()

    def _chargeback_executed(self, event):
        raise NotImplementedError()

    def _refund_created(self, event):
        raise NotImplementedError()

    def _refund_succeeded(self, event):
        raise NotImplementedError()

    def _refund_failed(self, event):
        raise NotImplementedError()

    def _subscription_created(self, event):
        self.object = BackendSubscription.objects.get(external_id=event['event_resource']['subscription']['id'])
        self.backend_event.content_object = self.object
        self.object.status = 'in_progress'

    def _subscription_updated(self, event):
        raise NotImplementedError()

    def _subscription_deleted(self, event):
        self.object.subscription.deleted = True
        self.object.subscription.save()
        self.object.canceled_at = event['event_resource']['subscription']['canceled_at']

    def _subscription_succeeded(self, event):
        self._subscription_created(event)
        self.object.status = 'paid'

    def _subscription_failed(self, event):
        self._subscription_created(event)
        self.object.status = 'failed'

    def _transaction_created(self, event):
        raise NotImplementedError()

    def _transaction_succeeded(self, event):
        raise NotImplementedError()

    def _transaction_failed(self, event):
        raise NotImplementedError()

    def _invoice_available(self, event):
        raise NotImplementedError()

    def _payout_transferred(self, event):
        raise NotImplementedError()

    def _app_merchant_activated(self, event):
        raise NotImplementedError()

    def _app_merchant_deactivated(self, event):
        raise NotImplementedError()

    def _app_merchant_rejected(self, event):
        raise NotImplementedError()

    def _client_updated(self, event):
        raise NotImplementedError()

    def _app_merchant_app_disabled(self, event):
        raise NotImplementedError()



from django.conf import settings


ABO_BACKENDS = getattr(settings, 'ABO_BACKENDS', tuple())
ABO_DEFAULT_BACKEND = getattr(settings, 'ABO_DEFAULT_BACKEND', '')
ABO_BACKENDS_SETTINGS = getattr(settings, 'ABO_BACKENDS_SETTINGS', dict())

SUBSCRIPTION_MODEL = getattr(settings, 'SUBSCRIPTION_MODEL', 'abo.Subscription')
PLAN_MODEL = getattr(settings, 'PLAN_MODEL', 'abo.Plan')



import json
import uuid
import logging

from django.http import HttpResponse, HttpResponseBadRequest
from django.core.urlresolvers import reverse
from django.views.decorators.csrf import csrf_exempt
from django.views.decorators.http import require_POST
from django.views.generic import FormView

from . import PaymentProcessor
from .forms import PaymillForm

from abo.models import BackendEvent

logger = logging.getLogger(__name__)


@csrf_exempt
@require_POST
def webhook(request, secret):
    # TODO: compare secret

    logger.debug('Message received')

    # TODO: find out backend from request url
    backend = "abo.backends.paymill"
    message = json.loads(request.body)

    # TODO: find out livemode - it's not in the event json
    livemode = False

    logger.debug(message)

    if type(message) != dict or not message.get("event") or type(message["event"]) != dict or not message["event"].get("event_type"):
        logger.warning("no event_type in message")
        return HttpResponseBadRequest()

    backend_event = BackendEvent.objects.create(
        backend=backend,
        external_id=str(uuid.uuid4()),  # paymill provides no id for events
        event_type=message["event"]["event_type"],
        livemode=livemode,
        message=message
    )
    backend_event.process()

    # always return an empty page, HTTP 200
    return HttpResponse()


class PaymillView(FormView):
    form_class = PaymillForm
    template_name = "paymill.html"
    backend = 'abo.backends.paymill'

    def get_context_data(self, **kwargs):
        context = super(PaymillView, self).get_context_data(**kwargs)
        context['PAYMILL_PUBLIC_KEY'] = PaymentProcessor.get_backend_setting('PAYMILL_PUBLIC_KEY')
        return context

    def form_valid(self, form):
        self.subscription = form.subscription
        return super(PaymillView, self).form_valid(form)

    def get_success_url(self):
        return reverse('abo-success', kwargs={'pk': self.subscription.pk})



from datetime import datetime
import pymill


class Mockmill(object):
    def __init__(*args, **kwargs):
        pass

    def new_client(self, *args, **kwargs):
        return pymill.Client(
            id="client1",
            email=kwargs.get('email'),
            created_at=datetime.now()
        )

    def new_card(self, token, client):
        '''return a new card without checking the token'''
        return pymill.Payment(
            id="payment1",
            client=client,
            type="creditcard",
            card_type="mastercard",
            card_holder="hans horst",
            expire_month="01",
            expire_year="2017",
            last_4="1234",
            code="123",
            created_at=datetime.now()
        )

    def new_offer(*args, **kwargs):
        return pymill.Offer(
            id="offer1",
            name="offer1",
            amount=1911,
            interval="1 MONTH",
            trial_perioade_days=None
        )

    def new_subscription(self, *args, **kwargs):
        return pymill.Subscription(
            id="subscription1",
            offer="offer1",
            livemode=False,
            cancel_at_period_end=None,
            canceled_at=None,
            interval="month",
            payment="payment1",
            client="client1",
            next_capture_at=datetime.now(),
            created_at=datetime.now(),
            updated_at=datetime.now()
        )


class MockmillFailOffer(Mockmill):
    def new_offer(*args, **kwargs):
        json_data = {'exception': 'mocked exception', 'error': 'something went wrong'}
        raise Exception(json_data)



from django.db import models
from django.utils.translation import ugettext_lazy as _

from abo.fields import JSONField


class Event(models.Model):
    received_at = models.DateTimeField(_("received at"), auto_now_add=True, db_index=True)
    event_type = models.CharField(max_length=250)
    data = JSONField()



import logging
import pymill
from datetime import datetime

from django import forms
from django.utils.translation import ugettext_lazy as _
from django.core.exceptions import ValidationError

from abo.signals import payment_error
from abo import get_subscription_model, get_plan_model
from abo.models import BackendPayment, BackendPlan, BackendSubscription, BackendClient

from . import PaymentProcessor


Plan = get_plan_model()
Subscription = get_subscription_model()

logger = logging.getLogger(__name__)


class PaymillForm(forms.Form):
    token = forms.CharField(widget=forms.HiddenInput())
    plan = forms.ModelChoiceField(queryset=Plan.objects.all())
    quantity = forms.IntegerField()
    email = forms.EmailField(help_text=_('Receipts will be sent to this email address'))

    def __init__(self, *args, **kwargs):
        self.Pymill = pymill.Pymill
        return super(PaymillForm, self).__init__(*args, **kwargs)

    def set_pymill(self, pymill_class):
        '''IoC for testings'''
        self.Pymill = pymill_class

    def raise_paymill_validation_error(self, exception):
        if len(exception.args) and type(exception.args[0]) == dict:
            error_code = exception.args[0].get('exception', 'error')
            error_message = exception.args[0].get('error', 'Error')
        else:
            error_code = 'error'
            error_message = 'Error'

        payment_error.send(sender=self, error_code=error_code, exception=exception)

        # TODO: notify developers
        raise ValidationError(_("Payment failed. Our developers have been notified."), code=error_code)

    def clean(self):
        cleaned_data = super(PaymillForm, self).clean()

        token = cleaned_data.get('token')
        plan = cleaned_data.get('plan')
        quantity = cleaned_data.get('quantity')
        email = cleaned_data.get('email')

        if not token or not plan or not quantity or not email:
            return cleaned_data

        self.subscription = None

        paymill = self.Pymill(PaymentProcessor.get_backend_setting('PAYMILL_PRIVATE_KEY'))

        self.backend = PaymentProcessor.BACKEND

        logger.debug('create client and save to DB')

        try:
            client = paymill.new_client(email=email)
        except Exception as e:
            self.raise_paymill_validation_error(e)

        backend_client = BackendClient.objects.create(
            backend=self.backend,
            external_id=client.id,
            created_at=client.created_at,
            email=email
        )

        logger.debug('create payment (card/direct debit) and save to DB')

        try:
            # new_card() also works for direct debit
            payment = paymill.new_card(token=token, client=backend_client.external_id)
        except Exception as e:
            self.raise_paymill_validation_error(e)

        backend_payment = BackendPayment.objects.create(
            backend=self.backend,
            external_id=payment.id,
            client=backend_client,
            payment_type=payment.type,
            card_type=payment.card_type,
            holder=payment.card_holder or payment.holder,  # cc --> card_holder; dd --> holder
            expire_month=payment.expire_month,
            expire_year=payment.expire_year,
            last4=payment.last4,
            code=payment.code,
            account=payment.account,
            iban=payment.iban,
            bic=payment.bic,
            created_at=payment.created_at
            # updated_at=payment.updated_at
        )

        backend_plan = BackendPlan.objects.filter(
            backend=self.backend,
            plan=plan,
            quantity=quantity
        )

        if not backend_plan.exists():
            logger.debug('create offer on paymill and save it to DB')

            amount = plan.amount * quantity
            interval = "{} {}".format(
                plan.interval_count or 1,
                plan.interval)
            name = "{}, Quantity: {}".format(
                plan.name,
                quantity)

            try:
                offer = paymill.new_offer(
                    amount=amount,
                    interval=interval,
                    name=name,
                    currency='EUR'
                )
            except Exception as e:
                self.raise_paymill_validation_error(e)

            subscription_count = getattr(offer, 'subscription_count', {'active': None, 'inactive': None})

            backend_plan = BackendPlan.objects.create(
                backend=self.backend,
                external_id=offer.id,
                plan=plan,
                quantity=quantity,
                subscription_count_active=subscription_count.get('active'),
                subscription_count_inactive=subscription_count.get('inactive'),
                created_at=getattr(offer, 'created_at', datetime.now()),
                updated_at=getattr(offer, 'updated_at', datetime.now())
            )
        else:
            logger.debug('offer already created, using it')
            backend_plan = backend_plan[0]

        logger.debug('create subscription')

        self.subscription = Subscription.objects.create(
            plan=plan,
            quantity=quantity,
            currency='EUR',
            # LOL APP SPECIFIC FIELD!!?!?!
        )

        logger.debug('subscribe!')

        try:
            subscription = paymill.new_subscription(
                client=backend_client.external_id,
                offer=backend_plan.external_id,
                payment=backend_payment.external_id
            )
        except Exception as e:
            self.raise_paymill_validation_error(e)

        BackendSubscription.objects.create(
            backend=self.backend,
            external_id=subscription.id,
            subscription=self.subscription,
            backend_plan=backend_plan,
            backend_payment=backend_payment,
            client=backend_client,
            status='new',
            created_at=subscription.created_at,
            updated_at=subscription.updated_at,
            # next_capture_at=subscription.next_capture_at, # TODO
            # canceled_at=subscription.canceled_at # TODO
        )

        logger.info("successfuly subscribed %s to %s" % (backend_client, self.subscription))

        return cleaned_data



import json

from django.test import TestCase
from django.core.urlresolvers import reverse

from abo.factories import PlanFactory
from abo.backends.paymill.forms import PaymillForm
from abo.models import BackendSubscription

from .mockups import Mockmill


class EventTestCase(TestCase):
    def _create_subscription(self):
        self.plan = PlanFactory()
        self.formdata = {
            'token': 'xxx123',
            'plan': self.plan.id,
            'quantity': 1,
            'email': 'test@example.com'
        }
        form = PaymillForm(data=self.formdata)
        form.set_pymill(Mockmill)
        form.full_clean()

    def setUp(self):
        self._create_subscription()
        self.secret = "0" * 32
        self.webhook_url = reverse('abo-paymill-webhook', args=[self.secret, ])

    def test_successful_subscription(self):
        msg = {u'event': {u'event_resource': {u'transaction': {u'status': u'closed', u'created_at': 1393435183, u'description': u'Subscription#subscription1 Almost Free Extended 50% off, Quantity: 3', u'refunds': None, u'invoices': [], u'response_code': 20000, u'livemode': False, u'origin_amount': 110094, u'updated_at': 1393435184, u'preauthorization': None, u'app_id': None, u'currency': u'EUR', u'amount': u'110094', u'short_id': u'7357.7357.7357', u'client': {u'description': None, u'payment': [u'payment1'], u'created_at': 1393435182, u'updated_at': 1393435182, u'app_id': None, u'id': u'client1', u'email': u'admin@chatgrape.com', u'subscription': None}, u'fees': [], u'id': u'transaction1', u'is_fraud': False, u'payment': {u'expire_month': u'11', u'country': None, u'created_at': 1393435181, u'app_id': None, u'updated_at': 1393435183, u'card_type': u'mastercard', u'last4': u'0004', u'client': u'client1', u'type': u'creditcard', u'expire_year': u'2020', u'card_holder': u'l\xf6kasdf', u'id': u'payment1'}}, u'subscription': {u'trial_start': None, u'cancel_at_period_end': False, u'offer': {u'subscription_count': {u'active': u'1', u'inactive': 0}, u'name': u'Almost Free Extended 50% off, Quantity: 3', u'created_at': 1393435183, u'interval': u'3 YEAR', u'app_id': None, u'updated_at': 1393435183, u'currency': u'EUR', u'amount': 110094, u'trial_period_days': 0, u'id': u'offer1'}, u'canceled_at': None, u'created_at': 1393435183, u'livemode': False, u'updated_at': 1393435184, u'app_id': None, u'trial_end': None, u'client': {u'description': None, u'payment': [{u'expire_month': u'11', u'country': None, u'created_at': 1393435181, u'app_id': None, u'updated_at': 1393435183, u'card_type': u'mastercard', u'last4': u'0004', u'client': u'client1', u'type': u'creditcard', u'expire_year': u'2020', u'card_holder': u'l\xf6kasdf', u'id': u'payment1'}], u'created_at': 1393435182, u'updated_at': 1393435182, u'app_id': None, u'id': u'client1', u'email': u'admin@chatgrape.com', u'subscription': None}, u'next_capture_at': 1488129584, u'id': u'subscription1', u'payment': {u'expire_month': u'11', u'country': None, u'created_at': 1393435181, u'app_id': None, u'updated_at': 1393435183, u'card_type': u'mastercard', u'last4': u'0004', u'client': u'client1', u'type': u'creditcard', u'expire_year': u'2020', u'card_holder': u'l\xf6kasdf', u'id': u'payment1'}}}, u'created_at': 1393435184, u'event_type': u'subscription.succeeded', u'app_id': None}}

        self.client.post(self.webhook_url, json.dumps(msg), content_type="application/json")

        bs = BackendSubscription.objects.get(external_id='subscription1')
        self.assertEqual(bs.status, 'paid')

    def test_failed_subscription(self):
        msg = {u'event': {u'event_resource': {u'transaction': {u'status': u'failed', u'created_at': 1393529473, u'description': u'Subscription#subscription1 Product Extended 50% off, Quantity: 5', u'refunds': None, u'invoices': [], u'response_code': 40103, u'livemode': False, u'origin_amount': 16905, u'updated_at': 1393529473, u'preauthorization': None, u'app_id': None, u'currency': u'EUR', u'amount': u'16905', u'short_id': None, u'client': {u'description': None, u'payment': [u'payment1'], u'created_at': 1393529473, u'updated_at': 1393529473, u'app_id': None, u'id': u'client1', u'email': u'admin@chatgrape.com', u'subscription': None}, u'fees': [], u'id': u'transaction1', u'is_fraud': False, u'payment': {u'expire_month': u'4', u'country': None, u'created_at': 1393529472, u'app_id': None, u'updated_at': 1393529473, u'card_type': u'mastercard', u'last4': u'5100', u'client': u'client1', u'type': u'creditcard', u'expire_year': u'2020', u'card_holder': u'lol rofl', u'id': u'payment1'}}, u'subscription': {u'trial_start': None, u'cancel_at_period_end': False, u'offer': {u'subscription_count': {u'active': u'2', u'inactive': 0}, u'name': u'Product Extended 50% off, Quantity: 5', u'created_at': 1393529369, u'interval': u'2 YEAR', u'app_id': None, u'updated_at': 1393529369, u'currency': u'EUR', u'amount': 16905, u'trial_period_days': 0, u'id': u'offer_f056f0ad034952fd8992'}, u'canceled_at': None, u'created_at': 1393529473, u'livemode': False, u'updated_at': 1393529473, u'app_id': None, u'trial_end': None, u'client': {u'description': None, u'payment': [{u'expire_month': u'4', u'country': None, u'created_at': 1393529472, u'app_id': None, u'updated_at': 1393529473, u'card_type': u'mastercard', u'last4': u'5100', u'client': u'client1', u'type': u'creditcard', u'expire_year': u'2020', u'card_holder': u'lol rofl', u'id': u'payment1'}], u'created_at': 1393529473, u'updated_at': 1393529473, u'app_id': None, u'id': u'client1', u'email': u'admin@chatgrape.com', u'subscription': None}, u'next_capture_at': 1456601473, u'id': u'subscription1', u'payment': {u'expire_month': u'4', u'country': None, u'created_at': 1393529472, u'app_id': None, u'updated_at': 1393529473, u'card_type': u'mastercard', u'last4': u'5100', u'client': u'client1', u'type': u'creditcard', u'expire_year': u'2020', u'card_holder': u'lol rofl', u'id': u'payment1'}}}, u'created_at': 1393529473, u'event_type': u'subscription.failed', u'app_id': None}}

        self.client.post(self.webhook_url, json.dumps(msg), content_type="application/json")

        bs = BackendSubscription.objects.get(external_id='subscription1')
        self.assertEqual(bs.status, 'failed')



from django.conf.urls import patterns, url, include

from abo.utils import import_backend_modules


backend_specific_urls = []
for backend_name, urls in import_backend_modules('urls').items():
    simple_name = backend_name.split('.')[-1]
    backend_specific_urls.append(url(r'^%s/' % simple_name, include(urls)))


from .views import (
    CancelView,
    ChangeCardView,
    ChangePlanView,
    HistoryView,
    SubscribeView,
    SubscriptionSuccessView,
    SubscriptionFailureView
)

urlpatterns = patterns('',
    url(r"^subscribe/$", SubscribeView.as_view(), name="abo-subscribe"),
    url(r"^subscription/(?P<pk>\d+)/success/$", SubscriptionSuccessView.as_view(), name="abo-success"),
    url(r"^subscription/(?P<pk>\d+)/failure/$", SubscriptionFailureView.as_view(), name="abo-failure"),
    url(r"^change/card/$", ChangeCardView.as_view(), name="abo-change_card"),
    url(r"^change/plan/$", ChangePlanView.as_view(), name="abo-change_plan"),
    url(r"^cancel/$", CancelView.as_view(), name="abo-cancel"),
    url(r"^history/$", HistoryView.as_view(), name="abo-history"),
    *backend_specific_urls
)



# -*- coding: utf-8 -*-

import logging

from django.db import models
from django.utils.translation import ugettext_lazy as _
from django.core.urlresolvers import reverse
from django.contrib.contenttypes import generic
from django.contrib.contenttypes.models import ContentType

from . import settings, signals
from .utils import import_name
from .fields import JSONField
from .choices import *

logger = logging.getLogger(__name__)


class AbstractPlan(models.Model):
    """
    Represents a plan a customer can subscribe to.

    This should be subclassed; the subclass should be registered as
    PLAN_MODEL in the settings.
    """
    name = models.CharField(max_length=512)
    amount = models.IntegerField(_('amount in cents'))
    interval = models.CharField(max_length=10, choices=PLAN_INTERVAL_CHOICES)
    interval_count = models.SmallIntegerField()
    trial_period_days = models.IntegerField(null=True)
    visible = models.BooleanField(_('visible'), default=True)
    created_at = models.DateTimeField(_("created at"), auto_now_add=True, db_index=True)

    class Meta:
        abstract = True

    def __unicode__(self):
        return "%s: %.2f/%s %s" % (self.name, self.amount / 100, self.interval_count, self.interval)


class Plan(AbstractPlan):
    """ Default Plan Model. If you want to change this, check out AbstractPlan """
    class Meta(AbstractPlan.Meta):
        swappable = 'PLAN_MODEL'


class AbstractSubscription(models.Model):
    """
    Represents a subscription, created when a customer subscribes to a plan.

    This should be subclassed; the subclass should be registered as
    SUBSCRIPTION_MODEL in the settings
    """
    plan = models.ForeignKey(settings.PLAN_MODEL, verbose_name=_('Plan'))
    quantity = models.IntegerField()
    total = models.DecimalField(decimal_places=2, max_digits=8, default=0)
    currency = models.CharField(max_length=3, default='EUR')
    deleted = models.BooleanField(default=False)
    #TODO: generic foreign key "subscriber"

    class Meta:
        abstract = True

    def __init__(self, *args, **kwargs):
        subscription = super(AbstractSubscription, self).__init__(*args, **kwargs)
        signals.new_subscription.send(sender=None, subscription=subscription)
        return subscription

    def __unicode__(self):
        return "%s x '%s', %.2f %s" % (self.quantity, self.plan, self.total, self.currency)

    def save(self, *args, **kwargs):
        self.total = self.quantity * self.plan.amount / 100
        return super(AbstractSubscription, self).save(*args, **kwargs)


class Subscription(AbstractSubscription):
    """ Default Subscription Model. If you want to change this, check out AbstractSubscription """
    class Meta(AbstractSubscription.Meta):
        swappable = 'SUBSCRIPTION_MODEL'


class BackendModel(models.Model):
    """
    All Models that save backend specific information (BackendPlan, BackendSubscription) always need to save the id the payment gateway created for the object and the creation time.
        """
    backend = models.CharField(_("backend"), max_length=50, choices=BACKEND_CHOICES)
    external_id = models.CharField(max_length=64)
    created_at = models.DateTimeField(_("created at"), auto_now_add=True, db_index=True)

    class Meta:
        abstract = True


class BackendPayment(BackendModel):
    client = models.ForeignKey('BackendClient', null=True)
    payment_type = models.CharField(_("payment type"), max_length=50)
    message = JSONField()
    livemode = models.BooleanField(default=False)
    updated_at = models.DateTimeField(_("updated_at"), auto_now_add=True)
    holder = models.CharField(_("account holder"), max_length=200, blank=True, null=True)

    # for credit cards only
    expire_month = models.PositiveSmallIntegerField(_("expire month"), blank=True, null=True)
    expire_year = models.PositiveSmallIntegerField(_("expire year"), blank=True, null=True)
    last4 = models.CharField(_("last four digits"), max_length=4, blank=True, null=True)
    card_type = models.CharField(_("card type"), max_length=50, blank=True, null=True)

    # for debit accounts only
    code = models.CharField(_("Bank code"), max_length=50, blank=True, null=True)
    account = models.CharField(_("Bank account"), max_length=50, blank=True, null=True)
    iban = models.CharField(_("IBAN"), max_length=50, blank=True, null=True)
    bic = models.CharField(_("BIC"), max_length=50, blank=True, null=True)


class BackendPlan(BackendModel):
    """
    Holds all the information on a plan on a specific backend. A Plan can have multiple BackendPlans, if you use multiple backends or if one backend needs them for internal reasons (see the following comment).

    Comment on field `quantity`: this is needed for backends that don't support to subscribe to a "quantity" of plans.
    Example: You offer a plan for 1€/user/month. A user can subscribe to the plan with 5 users and pay 5€/month. this shouldn't change the plan but only the quantity, saved in the subscription model. Stripe supports this, Paymill doesn't. For Paymill we have to create a seperate BackendPlan for every quantity.
    """
    plan = models.ForeignKey(settings.PLAN_MODEL)
    quantity = models.IntegerField(blank=True, null=True)
    updated_at = models.DateTimeField(_("updated on"), auto_now_add=True)
    subscription_count_active = models.IntegerField(_("active subscribers"), null=True)
    subscription_count_inactive = models.IntegerField(_("inactive subscribers"), null=True)


class BackendSubscription(BackendModel):
    """
    Holds all the information on a subscription on a specific backend. A BackendSubscription is always connected to a single BackendPlan from the same backend.
    """
    subscription = models.ForeignKey(settings.SUBSCRIPTION_MODEL)
    backend_plan = models.ForeignKey(BackendPlan)
    backend_payment = models.ForeignKey(BackendPayment)
    client = models.ForeignKey('BackendClient', null=True)
    status = models.CharField(_("status"), max_length=20, choices=SUBSCRIPTION_STATUS_CHOICES, default='new', db_index=True)
    updated_at = models.DateTimeField(_("updated on"), auto_now_add=True)
    next_capture_at = models.DateTimeField(_("next capture on"), null=True)
    canceled_at = models.DateTimeField(_("canceled on"), null=True)

    def get_absolute_url(self):
        return reverse('abo-success',
                       args={'pk': self.pk})


class BackendEvent(BackendModel):
    event_type = models.CharField(max_length=250)
    livemode = models.BooleanField(default=False)
    message = JSONField()
    processed = models.BooleanField(default=False)
    content_type = models.ForeignKey(ContentType, null=True)
    object_id = models.PositiveIntegerField(null=True)
    content_object = generic.GenericForeignKey('content_type', 'object_id')

    def process(self):
        if not self.processed:
            backend = import_name(self.backend + ".events")
            processor = getattr(backend, "EventProcessor")
            processor(self).process()

            self.processed = True
            self.save()


class BackendClient(BackendModel):
    """
    Holds information on a client on a specific backend. The actual payment info is saved in BackendPayment. Paymill needs a client object, but other backends might not.
    """
    email = models.CharField(max_length=300, blank=True, null=True)
    description = models.CharField(max_length=300, blank=True, null=True)
    updated_at = models.DateTimeField(_("updated_at"), auto_now_add=True, db_index=True)

    def __unicode__(self):
        return u"%s (%s)" % (self.email or "No Email", self.backend)

# class BackendTransaction(BackendModel):
#     amount = models.IntegerField(_('amount in cents'))
#     origin_amount = models.IntegerField(_('used amount in cents'))  # other currency
#     currency = models.CharField(max_length=3, default='EUR')
#     status = models.CharField(_("status"), max_length=20, choices=TRANSACTION_STATUS_CHOICES, db_index=True)
#     livemode = models.BooleanField()
#     is_fraud = models.BooleanField()
#     payment = models.ForeignKey(BackendPayment)
#     updated_at = models.DateTimeField(_("updated at"))
#     response_code = models.CharField(_("response code"), max_length=20)
#     short_id = models.CharField(_("short id"), max_length=100)
#     fees = models.TextField(_("fees"))

from abo.signals import payment_error
from django.dispatch import receiver


@receiver(payment_error)
def payment_error_callback(sender, error_code, exception, **kwargs):
    logger.error(error_code)



import requests
import json

from django.test import TestCase
from django.core.urlresolvers import reverse
from django.contrib.auth import get_user_model

from abo.factories import PlanFactory
from abo.models import BackendSubscription
from abo import get_subscription_model

Subscription = get_subscription_model()
User = get_user_model()


class IntegrationTestcase(TestCase):
    '''
    This tests the url config and the view (not only the form) using
    the real paymill testing backend.
    So this should completely test the payment, except for the frontend
    '''

    def setUp(self):
        self.plan = PlanFactory()
        self.user = User.objects.create_user('test', 'test@example.com', 'testpassword')
        self.client.login(username='test', password='testpassword')
        self.assertEqual(self.client.session['_auth_user_id'], self.user.pk)
        self.formdata = {
            'token': 'xxx123',
            'plan': self.plan.id,
            'quantity': 1,
            'email': 'test@example.com'
        }
        self.paymilldata = {
            'transaction.mode': 'CONNECTOR_TEST',
            'channel.id': '611273627719438339e4f177bcd9415c',
            'response.url': 'https://test-tds.paymill.de/end.php?parentUrl=http%253A%252F%252Flocalhost%253A8000%252Fpayment%252Fpaymill%252Fsubscribe%252F&',
            'jsonPFunction': 'window.paymill.transport.paymillCallback4002331054',
            'account.number': '4111111111111111',
            'account.expiry.month': '01',
            'account.expiry.year': '2021',
            'account.verification': '123',
            'account.holder': 'Test Test'
        }

    def test_html_contains_token_input(self):
        response = self.client.post(
            reverse('abo-paymill-authorization'),
            {'text': ''})

        token_input = '<input id="id_token" name="token" type="hidden" />'

        self.assertInHTML(token_input, response.content)

    def test_fake_token(self):
        response = self.client.post(
            reverse('abo-paymill-authorization'), self.formdata)

        # assertFormError has to be fixed in django
        # self.assertFormError(response, form='PaymillForm', field='__all__', errors="Token not Found")
        form = response.context[0].get('form')

        # Important: no subscription and backendsubscription have been created
        self.assertEquals(Subscription.objects.count(), 0)
        self.assertEquals(BackendSubscription.objects.count(), 0)

        self.assertTrue('Payment failed. Our developers have been notified.' in form.errors.get('__all__'))

    def test_fake_plan(self):
        self.formdata['plan'] = 'xxxx'
        response = self.client.post(
            reverse('abo-paymill-authorization'), self.formdata)
        form = response.context[0].get('form')

        # Important: no subscription and backendsubscription have been created
        self.assertEquals(Subscription.objects.count(), 0)
        self.assertEquals(BackendSubscription.objects.count(), 0)

        self.assertTrue('Select a valid choice. That choice is not one of the available choices.' in form.errors.get('plan'))

    def test_successful_payment(self):
        # request token from paymill test API. this is usually done in the
        # frontend with JS, but this is not a frontend test

        r1 = requests.get('https://test-token.paymill.de/', params=self.paymilldata)

        # ugly: strip away JS callback function and parse to json

        r1_cleaned = r1.text[r1.text.index('(') + 1:-1]
        r1_json = json.loads(r1_cleaned)

        # check if our request was processed and if we got a token

        self.assertEquals(r1_json['transaction']['processing']['result'], 'ACK')
        self.formdata['token'] = r1_json['transaction']['identification']['uniqueId']
        self.assertIn('tok_', self.formdata['token'])

        # now that we have the token, make a request to our own paymill view

        r2 = self.client.post(reverse('abo-paymill-authorization'), self.formdata)

        self.assertIn('/success/', r2.url)



from django.core.management import BaseCommand

from abo.utils import import_backend_modules, import_name


class Command(BaseCommand):
    def handle(self, *args, **kwargs):
        backends = import_backend_modules()

        for name, cls_name in backends.iteritems():
            try:
                webhook_module = import_name(name + ".webhooks")

            except AttributeError:
                # this module has no webhooks, just ignore it.
                print "Backend '%s' has no webhooks" % name
                continue

            print "Registering webhook for backend '%s'" % name
            webhook = webhook_module.Webhook()
            webhook.init_webhook()
            print "Webhook URL: %s" % webhook.get_url()



from django.test import TestCase

from abo.factories import PlanFactory
from abo.backends.paymill.forms import PaymillForm
from abo.models import BackendPayment, BackendPlan, BackendSubscription
from abo import get_subscription_model

from .mockups import Mockmill, MockmillFailOffer

Subscription = get_subscription_model()


class FormHandlingTestCase(TestCase):
    def setUp(self):
        self.plan = PlanFactory()
        self.formdata = {
            'token': 'xxx123',
            'plan': self.plan.id,
            'quantity': 1,
            'email': 'test@example.com'
        }

    def test_failing_paymill_offer(self):
        backend_offer_count_before = BackendPlan.objects.count()

        form = PaymillForm(data=self.formdata)
        form.set_pymill(MockmillFailOffer)
        form.full_clean()

        self.assertEquals(BackendPayment.objects.filter(external_id="payment1").count(), 1)
        self.assertEquals(BackendPlan.objects.count(), backend_offer_count_before)

        # Important: no subscription and backendsubscription have been created
        self.assertEquals(Subscription.objects.count(), 0)
        self.assertEquals(BackendSubscription.objects.count(), 0)

        self.assertTrue('Payment failed. Our developers have been notified.' in form.errors.get('__all__'))

    def test_successful_payment(self):
        form = PaymillForm(data=self.formdata)
        form.set_pymill(Mockmill)
        form.full_clean()

        self.assertEquals(BackendPayment.objects.filter(external_id="payment1").count(), 1)
        self.assertEquals(BackendPlan.objects.filter(external_id="offer1").count(), 1)
        self.assertEquals(BackendSubscription.objects.filter(external_id="subscription1").count(), 1)

        self.assertEquals(form.subscription, BackendSubscription.objects.get(external_id="subscription1").subscription)
        self.assertEquals(form.subscription.plan, BackendPlan.objects.get(external_id="offer1").plan)



from django.conf.urls import patterns, url
from .views import PaymillView, webhook


urlpatterns = patterns('',
    url(r'subscribe/$', PaymillView.as_view(), name='abo-paymill-authorization'),
    url(r'^webhook/(?P<secret>\w{32})$', webhook, name='abo-paymill-webhook')
)



from django.test import TestCase

from abo.models import BackendEvent


class BackendEventTestCase(TestCase):
    def test_event_processor_add_object(self):
        event = BackendEvent.objects.create(
            backend="abo.tests.mocked_backend_for_events",
            livemode=False,
            message={"method": "new_plan", "name": "test123"}
        )

        # processor should create new plan
        event.process()

        self.assertTrue(event.processed)
        self.assertEqual(event.content_object.name, "test123")

    def test_event_processor_fail(self):
        event = BackendEvent.objects.create(
            backend="abo.tests.mocked_backend_for_events",
            livemode=False,
            message={"method": "fail", "name": "test123"}
        )

        # processor should fail
        self.assertRaises(NotImplementedError, event.process)

        self.assertFalse(event.processed)
        self.assertEquals(event.object_id, None)



from . import settings
from django.core.exceptions import ImproperlyConfigured


def get_plan_model():
    """
    Returns the Plan model that is active in this project.
    """
    from django.db.models import get_model

    try:
        app_label, model_name = settings.PLAN_MODEL.split('.')
    except ValueError:
        raise ImproperlyConfigured("PLAN_MODEL must be of the form 'app_label.model_name'")
    plan_model = get_model(app_label, model_name)
    if plan_model is None:
        raise ImproperlyConfigured("PLAN_MODEL refers to model '%s' that has not been installed" % settings.PLAN_MODEL)
    return plan_model


def get_subscription_model():
    """
    Returns the Subscription model that is active in this project.
    """
    from django.db.models import get_model

    try:
        app_label, model_name = settings.SUBSCRIPTION_MODEL.split('.')
    except ValueError:
        raise ImproperlyConfigured("SUBSCRIPTION_MODEL must be of the form 'app_label.model_name'")
    subscription_model = get_model(app_label, model_name)
    if subscription_model is None:
        raise ImproperlyConfigured("SUBSCRIPTION_MODEL refers to model '%s' that has not been installed" % settings.SUBSCRIPTION_MODEL)
    return subscription_model



from abo.factories import PlanFactory


class events(object):
    class EventProcessor:
        def __init__(self, backend_event):
            self.backend_event = backend_event

        def process(self):
            if self.backend_event.message.get("method") is "new_plan":
                plan = PlanFactory(name=self.backend_event.message.get("name"))
                self.backend_event.content_object = plan
                self.backend_event.save()
            else:
                raise NotImplementedError(self.backend_event.message.get("method"))



from django.core.urlresolvers import reverse
from django.utils.translation import ugettext_lazy as _

from abo.backends import PaymentProcessorBase

default_app_config = 'abo.backends.paymill.apps.PaymillConfig'


class PaymentProcessor(PaymentProcessorBase):
    BACKEND = 'abo.backends.paymill'
    BACKEND_NAME = _('Paymill')
    BACKEND_ACCEPTED_CURRENCY = ('EUR', 'CZK', 'DKK', 'HUF', 'ISK', 'ILS',
                                 'LVL', 'CHF', 'NOK', 'PLN', 'SEK', 'TRY',
                                 'GBP', )

    @classmethod
    def get_gateway_url(cls, request):
        return reverse('abo-paymill-authorization'), "GET", {}
